@@ -78,6 +78,14 @@ impl FromStr for Fish<PublicKeyAlgorithm> {
             "openpgp-rsa" |
             "oid.1.2.840.113549.1.1.1" =>
                 Ok(Fish(PublicKeyAlgorithm::RSAEncryptSign)),
+            "dsa" | "openpgp-dsa" =>
+                Ok(Fish(PublicKeyAlgorithm::DSA)),
+            "elg" |
+            "openpgp-elg" =>
+                Ok(Fish(PublicKeyAlgorithm::ElGamalEncrypt)),
+            "ecdh" => Ok(Fish(PublicKeyAlgorithm::ECDH)),
+            "eddsa" => Ok(Fish(PublicKeyAlgorithm::EdDSA)),
+            "ecdsa" => Ok(Fish(PublicKeyAlgorithm::ECDSA)),
             _ => {
                 if let Ok(o) = u8::from_str(s) {
                     Ok(Fish(o.into()))
@@ -131,12 +139,6 @@ impl FromStr for Fish<Curve> {
 }
 
 impl fmt::Display for Fish<PublicKeyAlgorithmAndSize> {
-    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        Fish(&self.0).fmt(f)
-    }
-}
-
-impl fmt::Display for Fish<&PublicKeyAlgorithmAndSize> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         use PublicKeyAlgorithmAndSize::*;
         match &self.0 {
