@@ -521,29 +521,7 @@ impl Record<'_> {
                         for n in notations.iter().rev() {
                             writeln!(w, "   Signature notation: {}={}",
                                      n.name(),
-                                     if n.flags().human_readable() {
-                                         String::from_utf8_lossy(n.value())
-                                             .to_string()
-                                     } else {
-                                         let preview: String =
-                                             n.value().iter().take(19)
-                                             .map(|b|
-                                                  char::from_u32((*b).into())
-                                                  .filter(|c| *c == ' '
-                                                          || c.is_ascii_graphic())
-                                                  .unwrap_or('?'))
-                                             .collect();
-
-                                         format!("[ not human readable \
-                                                  ({} bytes: {}{}) ]",
-                                                 n.value().len(),
-                                                 preview,
-                                                 if n.value().len() > 20 {
-                                                     "..."
-                                                 } else {
-                                                     ""
-                                                 })
-                                     })?;
+                                     crate::verify::format_notation_value(n))?;
                         }
                     }
                 }
