@@ -130,14 +130,11 @@ pub fn cmd_implicit(config: &crate::Config, args: &[String])
                     .map(|r| r.map(|c| Arc::new(LazyCert::from(c))))
                     .collect::<Result<Vec<_>>>()?;
             crate::list_keys::list_keys(
-                config, certs.into_iter(), false, false, io::stdout())
+                config, certs.into_iter(), false, false,
+                crate::io::stdout())
         },
         Some(InlineVerify) => {
-            let mut sink = if let Some(name) = config.outfile() {
-                utils::create(config, name)?
-            } else {
-                Box::new(io::sink())
-            };
+            let mut sink = config.sink()?;
             let helper = crate::verify::VHelper::new(config, 1);
             let mut v = VerifierBuilder::from_reader(input)?
                 .with_policy(config.policy(), config.now(), helper)?;
@@ -346,7 +343,7 @@ pub fn cmd_generate_revocation(config: &crate::Config, args: &[String])
     let q = args[0].parse()?;
     let always = trust::Always::default();
     let vtm = always.with_policy(config, Some(config.now()))?;
-    let certs = config.lookup_certs_with(vtm.as_ref(), &q, true)?;
+    let certs = vtm.lookup(&q)?;
 
     if certs.is_empty() {
         return Err(anyhow::anyhow!("secret key \"{}\" not found", q));
@@ -451,12 +448,7 @@ Your decision?"))?.to_lowercase().as_str()
             .set_reason_for_revocation(reason, description.as_bytes())?
             .build(&mut primary_signer, &cert, None)?;
 
-        let sink = if let Some(name) = config.outfile() {
-            utils::create(config, name)?
-        } else {
-            Box::new(io::stdout())
-        };
-
+        let sink = config.sink()?;
         let mut message = Message::new(sink);
 
         if config.armor {
@@ -488,11 +480,7 @@ pub fn cmd_enarmor(config: &crate::Config, args: &[String])
 
     let filename = args.get(0).cloned().unwrap_or_else(|| "-".into());
     let mut source = utils::open(config, &filename)?;
-    let sink = if let Some(name) = config.outfile() {
-        utils::create(config, name)?
-    } else {
-        Box::new(io::stdout())
-    };
+    let sink = config.sink()?;
 
     let mut sink = Writer::with_headers(
         sink, Kind::File,
@@ -510,11 +498,7 @@ pub fn cmd_dearmor(config: &crate::Config, args: &[String])
 
     let filename = args.get(0).cloned().unwrap_or_else(|| "-".into());
     let source = utils::open(config, &filename)?;
-    let mut sink = if let Some(name) = config.outfile() {
-        utils::create(config, name)?
-    } else {
-        Box::new(io::stdout())
-    };
+    let mut sink = config.sink()?;
 
     let mut source = Reader::from_reader(source, ReaderMode::Tolerant(None));
     std::io::copy(&mut source, &mut sink)?;
@@ -544,7 +528,7 @@ pub fn cmd_export_ownertrust(config: &crate::Config, args: &[String])
         return Err(anyhow::anyhow!("Expected no arguments, got some"));
     }
 
-    config.trustdb.export_ownertrust(&mut std::io::stdout())?;
+    config.trustdb.export_ownertrust(&mut crate::io::stdout())?;
     Ok(())
 }
 
