@@ -113,6 +113,9 @@ pub trait Common<'store> {
     /// Returns the output file.
     fn outfile(&self) -> Option<&String>;
 
+    /// Opens and returns the output stream.
+    fn sink(&self) -> Result<Box<dyn std::io::Write + Send + Sync>>;
+
     /// Returns the policy.
     fn policy(&self) -> &GPGPolicy;
 
