@@ -1,4 +1,5 @@
 use std::{
+    collections::HashMap,
     io,
 };
 
@@ -41,6 +42,7 @@ use crate::{
     common::Common,
     compliance::Compliance,
     error_codes,
+    list_keys::KeyInfoProxy,
     status::{
         NoDataReason,
         Status,
@@ -68,11 +70,7 @@ pub fn cmd_decrypt(config: &crate::Config, args: &[String])
     // XXX: Currently, there is no nice way to disable armoring when
     // using the streaming decryptor.
 
-    let sink = if let Some(name) = config.outfile() {
-        utils::create(config, name)?
-    } else {
-        Box::new(io::stdout())
-    };
+    let sink = config.sink()?;
 
     let r = if config.unwrap_encryption {
         decrypt_unwrap(config, Ok(message), Ok(sink))
@@ -429,8 +427,22 @@ impl<'a, 'store> DHelper<'a, 'store> {
             }
         }
 
-        let mut agent = self.config.connect_agent().await?;
-        let secret_keys = agent.list_keys().await.unwrap_or_default();
+        let mut secret_keys = KeyInfoProxy::empty();
+        async fn connect(config: &crate::Config<'_>,
+                         secret_keys: &mut KeyInfoProxy) -> Result<()>
+        {
+            if secret_keys.agent().is_some() {
+                return Ok(())
+            }
+
+            *secret_keys = KeyInfoProxy::connect(config).await?;
+            Ok(())
+        }
+
+        // For decrypting PKESKs with anonymous recipients, we build
+        // up a list of keys where we have secrets for.  We do that
+        // lazily, because it is expensive.
+        let mut keys_with_secrets = None;
 
         // First, try public key encryption.
         let mut success = None;
@@ -467,7 +479,99 @@ impl<'a, 'store> DHelper<'a, 'store> {
 
             let keyid = pkesk.recipient();
             if keyid.is_wildcard() {
-                continue; // XXX
+                if self.config.skip_hidden_recipients {
+                    *error = Some(error_codes::Error::GPG_ERR_NO_SECKEY);
+                    continue;
+                }
+
+                connect(self.config, &mut secret_keys).await?;
+
+                // This is the expensive part.  Iterate over all certs
+                // to find the ones with secret keys.
+                if keys_with_secrets.is_none() {
+                    // At least do it just once.
+                    let mut with_secrets = HashMap::new();
+
+                    for cert in self.config.keydb().certs()
+                        .filter(|c| c.to_cert()
+                                .map(|c| c.exportable()).unwrap_or(false))
+                    {
+                        let vcert = match cert.with_policy(self.config.policy(),
+                                                           self.config.now()) {
+                            Ok(c) => c,
+                            Err(_) => continue,
+                        };
+
+                        for key in vcert.keys()
+                            .for_transport_encryption()
+                            .for_storage_encryption()
+                        {
+                            if secret_keys.lookup_by_key(key.key())
+                                .await.is_some()
+                            {
+                                with_secrets.insert(key.key().clone(),
+                                                    cert.clone());
+                            }
+                        }
+                    }
+
+                    keys_with_secrets = Some(with_secrets);
+                }
+
+                // Now try every key.
+                for (key, cert) in keys_with_secrets.as_ref()
+                    .expect("computed above")
+                    .iter()
+                    .filter(|(k, _)| k.pk_algo() == pkesk.pk_algo())
+                {
+                    let vcert = match cert.with_policy(self.config.policy(),
+                                                       self.config.now()) {
+                        Ok(c) => c,
+                        Err(_) => continue,
+                    };
+
+                    let mut pair = secret_keys.agent_mut().expect("connected")
+                        .keypair(&key)?
+                        .with_cert(&vcert);
+
+                    // See if we have a static password to loop back to the
+                    // agent.
+                    if let (crate::gpg_agent::PinentryMode::Loopback, Some(p)) =
+                        (&self.config.pinentry_mode,
+                         self.config.static_passphrase.borrow().as_ref())
+                    {
+                        pair = pair.with_password(p.clone());
+                    }
+
+                    self.config.warn(format_args!(
+                        "anonymous recipient; trying secret key {} ...",
+                        key.keyid()));
+
+                    // GnuPG emits this line twice in `get_session_key`.
+                    // First to get the secret key using `get_seckey`, then
+                    // once more in `get_it`.
+                    for _ in 0..2 {
+                        self.config.status().emit(
+                            Status::KeyConsidered {
+                                fingerprint: cert.fingerprint(),
+                                not_selected: false,
+                                all_expired_or_revoked: false,
+                            })?;
+                    }
+
+                    if let Ok(r) = self.try_decrypt(
+                        cert.to_cert()?, pkesk, sym_algo, pair, decrypt).await
+                    {
+                        // Success!
+                        self.config.warn(format_args!(
+                            "okay, we are the anonymous recipient."));
+                        success = Some(r);
+                        break;
+                    }
+                }
+
+                // Try next PKESK.
+                continue;
             }
             let handle = KeyHandle::from(keyid);
 
@@ -509,7 +613,8 @@ impl<'a, 'store> DHelper<'a, 'store> {
                 },
             };
 
-            if secret_keys.lookup_by_key(key.key()).is_none() {
+            connect(self.config, &mut secret_keys).await?;
+            if secret_keys.lookup_by_key(key.key()).await.is_none() {
                 *error = Some(error_codes::Error::GPG_ERR_NO_SECKEY);
                 continue;
             }
@@ -531,7 +636,8 @@ impl<'a, 'store> DHelper<'a, 'store> {
             }
 
             // And just try to decrypt it using the agent.
-            let mut pair = agent.keypair(&key)?
+            let mut pair = secret_keys.agent_mut().expect("connected")
+                .keypair(&key)?
                 .with_cert(&vcert);
 
             // See if we have a static password to loop back to the
@@ -669,9 +775,10 @@ impl<'a, 'store> DHelper<'a, 'store> {
                 self.config.status().emit(Status::InquireMaxLen(100))?;
             }
 
+            connect(self.config, &mut secret_keys).await?;
             let p =
                 self.config.get_passphrase(
-                    &mut agent,
+                    secret_keys.agent_mut().expect("connected"),
                     &cacheid, &error, None, None, false, 0, false, false,
                     |p| {
                         let info = String::from_utf8_lossy(&p);
@@ -698,7 +805,7 @@ impl<'a, 'store> DHelper<'a, 'store> {
             error = Some("Decryption failed".to_string());
             if let Some(cacheid) = &cacheid {
                 // Make gpg-agent forget the bad passphrase.
-                agent.forget_passphrase(
+                secret_keys.agent_mut().expect("connected").forget_passphrase(
                     &cacheid,
                     |info| {
                         let info = String::from_utf8_lossy(&info);
