@@ -22,7 +22,6 @@ use crate::{
     common::{Common, Query},
     filter,
     status::{Status, ExportResult},
-    utils,
 };
 
 /// Controls export operations.
@@ -169,11 +168,7 @@ pub fn cmd_export(config: &mut crate::Config, args: &[String],
     // We collect stats for the final EXPORT_RES status line.
     let mut s = ExportResult::default();
 
-    let mut sink = if let Some(name) = config.outfile() {
-        utils::create(config, name)?
-    } else {
-        Box::new(std::io::stdout())
-    };
+    let mut sink = config.sink()?;
 
     let mut message = Message::new(&mut sink);
     if config.armor && ! config.export_options.dane {
