@@ -1136,7 +1136,7 @@ fn add_primary_key_metadata(builder: SignatureBuilder,
     }
 
     if let Some(ks) = preferred_keyserver {
-        builder = builder.set_preferred_key_server(ks.url())?;
+        builder = builder.set_preferred_key_server(&ks.url().to_string())?;
     }
 
     if let Some(r) = revoker {
