@@ -325,7 +325,7 @@ pub async fn do_import_cert(config: &mut crate::Config<'_>,
                     config,
                     vec![key.clone()].into_iter(),
                     true, false, false, false, false,
-                    std::io::stdout()).await?;
+                    crate::io::stdout()).await?;
             } else if for_migration {
                 // Be quiet when migrating.
             } else {
@@ -435,7 +435,7 @@ pub async fn do_import_cert(config: &mut crate::Config<'_>,
             list_keys::async_list_keys(
                 config, vec![key.clone()].into_iter(),
                 true, false, false, false, false,
-                std::io::stdout()).await?;
+                crate::io::stdout()).await?;
         } else if for_migration {
             // Be quiet when migrating.
         } else {
