@@ -33,6 +33,8 @@ use cert_store::Store;
 use cert_store::StoreUpdate;
 use cert_store::store::UserIDQueryParams;
 
+use openpgp_cert_d::CertD;
+
 use sequoia_wot as wot;
 
 use crate::{
@@ -50,6 +52,12 @@ pub struct KeyDB<'a> {
     // store.
     overlay: Result<Overlay<'a>, cert_store::store::Certs<'a>>,
 
+    /// An overlay for metadata.
+    ///
+    /// For example, the trust database overlay is stored there.  We
+    /// need this even if the default keyrings are disabled.
+    metadata_overlay: Option<CertD>,
+
     initialized: bool,
 }
 
@@ -121,6 +129,7 @@ impl<'store> KeyDB<'store> {
             resources: Vec::default(),
             initialized: false,
             overlay: Err(cert_store::store::Certs::empty()),
+            metadata_overlay: None,
         }
     }
 
@@ -133,20 +142,18 @@ impl<'store> KeyDB<'store> {
     }
 
     #[allow(dead_code)]
-    pub fn add_resource<U>(&mut self,
-                           home_dir: &Path,
-                           url: U,
-                           read_only: bool,
-                           default: bool)
-                           -> Result<()>
-    where
-        U: AsRef<str>,
+    pub fn add_resource(&mut self,
+                        home_dir: &Path,
+                        mut url: &str,
+                        gnupghome_relative: Option<bool>,
+                        read_only: bool,
+                        default: bool)
+                        -> Result<()>
     {
         tracer!(TRACE, "KeyDB::add_resource");
-        t!("home_dir {:?}, url {:?}, read_only {:?}, default {:?}",
-           home_dir, url.as_ref(), read_only, default);
+        t!("home_dir {:?}, url {:?}, gnupghome_relative: {:?}, read_only {:?}, default {:?}",
+           home_dir, url, gnupghome_relative, read_only, default);
 
-        let mut url = url.as_ref();
         let mut kind = None;
         let create = ! read_only && self.resources.is_empty();
 
@@ -161,12 +168,21 @@ impl<'store> KeyDB<'store> {
             url = &url[13..];
         }
 
-        // Expand tildes.
-        let mut path = PathBuf::from(shellexpand::tilde(url).as_ref());
+        // If there is no path separator in the URL, treat it as
+        // relative to the GNUPGHOME.  GnuPG first checks whether
+        // there is a path separator, then does the tilde expansion.
+        let gnupghome_relative = gnupghome_relative.unwrap_or_else(|| {
+            let p = PathBuf::from(url);
+            p.iter().count() == 1
+        });
 
-        if ! path.is_absolute() {
-            path = home_dir.join(path);
-        }
+        // Canonicalize the path.
+        let mut path = if gnupghome_relative {
+            home_dir.join(url)
+        } else {
+            // Expand tildes.
+            PathBuf::from(shellexpand::tilde(url).as_ref())
+        };
         t!("abolute path: {:?}", path);
 
         if kind.is_none() {
@@ -291,20 +307,48 @@ impl<'store> KeyDB<'store> {
 
     /// Adds a writable pgp-cert-d overlay to the resources, if not
     /// already in place.
+    ///
+    /// Implicitly adds the metadata overlay.
     pub fn add_certd_overlay(&mut self, path: &Path) -> Result<()> {
         tracer!(TRACE, "KeyDB::add_certd_overlay");
+        t!("Configuring certd overlay to use {}", path.display());
         if self.overlay.is_ok() {
             t!("CertD overlay already configured.");
             return Ok(());
         }
 
         self.overlay = Ok(Overlay::new(path)?);
+        self.metadata_overlay = Some(CertD::with_base_dir(path)?);
         Ok(())
     }
 
-    /// Gets the writable pgp-cert-d overlay.
-    pub fn get_certd_overlay(&self) -> Result<&Overlay<'store>> {
-        self.overlay.as_ref().map_err(|_| anyhow::anyhow!("No overlay added"))
+    /// Adds a writable pgp-cert-d overlay to the resources, if not
+    /// already in place.
+    pub fn add_metadata_overlay(&mut self, path: &Path) -> Result<()> {
+        tracer!(TRACE, "KeyDB::add_metadata_overlay");
+        t!("Configuring metadata overlay to use {}", path.display());
+        if self.metadata_overlay.is_some() {
+            t!("Metadata overlay already configured.");
+            return Ok(());
+        }
+
+        self.metadata_overlay = Some(CertD::with_base_dir(path)?);
+        Ok(())
+    }
+
+    /// Gets the writable pgp-cert-d overlay for use as a cert store.
+    ///
+    /// If `--no-default-keyring` has been given, this will fail.
+    pub fn certd_overlay(&self) -> Result<&Overlay<'store>> {
+        self.overlay.as_ref()
+            .map_err(|_| anyhow::anyhow!("No certd overlay added"))
+    }
+
+    /// Gets the writable pgp-cert-d overlay for use as metadata
+    /// store.
+    pub fn metadata_overlay(&self) -> Result<&CertD> {
+        self.metadata_overlay.as_ref()
+            .ok_or_else(|| anyhow::anyhow!("No metadata overlay added"))
     }
 
     // Initialize a keyring.
