@@ -144,6 +144,8 @@ impl KeyserverOptions {
     /// All import options and all export options are valid keyserver
     /// options too.
     pub fn parse(&mut self, s: &str) -> Result<()> {
+        // XXX: This doesn't look like the right way to chain the
+        // parsers...
         if let Ok(()) = argparse::options::parse(&Self::OPTS, s, self) {
             return Ok(());
         }
@@ -177,6 +179,14 @@ pub fn cmd_refresh_keys(config: &mut crate::Config, args: &[String])
     rt.block_on(keyserver_import(config, args, true))
 }
 
+/// Dispatches the --send-keys command.
+pub fn cmd_send_keys(config: &mut crate::Config, args: &[String])
+			-> Result<()>
+{
+    let rt = tokio::runtime::Runtime::new()?;
+    rt.block_on(keyserver_export(config, args))
+}
+
 /// Dispatches the --receive-keys and --refresh-keys commands.
 async fn keyserver_import(config: &mut crate::Config<'_>, args: &[String],
 			  refresh_keys: bool)
@@ -226,7 +236,7 @@ async fn keyserver_import(config: &mut crate::Config<'_>, args: &[String],
                 .for_url(k.url())?
                 .build()?;
 
-	    net::KeyServer::with_client(k.url(), c)
+	    net::KeyServer::with_client(&k.url().to_string(), c)
 	})
 	.collect::<Result<Vec<_>>>()?;
 
@@ -326,6 +336,132 @@ async fn importer(config: &mut crate::Config<'_>,
     Ok(())
 }
 
+/// Implements the --send-keys command.
+async fn keyserver_export(config: &mut crate::Config<'_>, args: &[String])
+			  -> Result<()>
+{
+    tracer!(TRACE, "keyserver::keyserver_export");
+
+    let handles: Vec<Query> =
+	args.iter()
+	.filter_map(|a| match a.parse() {
+	    Ok(Query::Key(h)) => Some(Query::Key(h)),
+            Ok(Query::ExactKey(h)) => Some(Query::ExactKey(h)),
+	    Ok(_) => {
+		config.error(format_args!(
+		    "{:?} not a key ID: skipping", a));
+		None
+	    },
+            Err(e) => {
+		config.error(format_args!(
+		    "{:?} not a key ID, skipping: {}", a, e));
+		None
+	    }
+	})
+	.collect();
+
+    if handles.is_empty() {
+        // Return success if the list is empty.
+        return Ok(());
+    }
+
+    let certs = handles.into_iter()
+        .filter_map(|h| match config.lookup_certs(&h) {
+            Ok(candidates) => match candidates.len() {
+                0 => {
+                    config.warn(format_args!(
+                        "skipped \"{}\": {}", h, "not found"));
+                    None
+                },
+                1 => candidates.into_iter().next().map(|(_, c)| (h, c)),
+                _ => {
+                    config.warn(format_args!(
+                        "skipped \"{}\": {}", h,
+                        "matches more than one cert"));
+                    None
+                },
+            },
+            Err(e) => {
+                config.warn(format_args!("skipped \"{}\": {}", h, e));
+                None
+            },
+        })
+    // Convert to certs (though this shouldn't fail, really).
+        .filter_map(|(h, c)| match c.to_cert() {
+            Ok(c) => Some(c.clone()),
+            Err(e) => {
+                config.warn(format_args!("skipped \"{}\": {}", h, e));
+                None
+            },
+        })
+    // XXX: Honor keyserver and export options once we implement some.
+        .collect::<Vec<_>>();
+
+    let servers =
+	config.keyserver.iter().map(|k| {
+            t!("Using server {}", k.url());
+
+	    let c = config.make_http_client()
+                .for_url(k.url())?
+                .build()?;
+
+	    net::KeyServer::with_client(&k.url().to_string(), c)
+	})
+	.collect::<Result<Vec<_>>>()?;
+
+    // Concurrently send all certs to all servers.
+    stream::iter(certs.iter())
+        .flat_map_unordered(None, |cert| {
+            let servers = &servers;
+	    stream::iter(servers)
+		.map(move |server| server.send(cert))
+        })
+        .buffer_unordered(CONCURRENT_REQUESTS)
+        .for_each(|r| async {
+            if let Err(e) = r {
+                config.warn(format_args!(
+                    "keyserver send failed: {}", e));
+            }
+        })
+        .await;
+
+    Ok(())
+}
+
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub struct KeyserverURL {
+    url: reqwest::Url,
+}
+
+impl KeyserverURL {
+    pub fn url(&self) -> &reqwest::Url {
+        &self.url
+    }
+}
+
+impl Default for KeyserverURL {
+    fn default() -> Self {
+        "hkps://keys.openpgp.org".parse().unwrap()
+    }
+}
+
+impl std::str::FromStr for KeyserverURL {
+    type Err = anyhow::Error;
+
+    fn from_str(s: &str) -> std::result::Result<Self, Self::Err> {
+        let s_;
+        let mut s = s;
+        if ! s.contains("://") {
+            s_ = format!("hkp://{}", s);
+            s = &s_;
+        }
+
+        Ok(Self {
+            url: s.parse()?,
+        })
+    }
+}
+
 #[derive(Default, Clone)]
 pub struct HttpClientBuilder {
     connect_timeout: Duration,
