@@ -1,10 +1,13 @@
 use std::{
+    borrow::Borrow,
     cell::OnceCell,
     collections::{
         BTreeMap,
         BTreeSet,
+        HashMap,
+        HashSet,
     },
-    io::{self, Write},
+    io::Write,
     sync::Arc,
 };
 
@@ -35,7 +38,7 @@ use cert_store::Store;
 use cert_store::store::{StoreError, UserIDQueryParams};
 
 use sequoia_gpg_agent::{
-    keyinfo::KeyInfoList,
+    keyinfo::KeyInfo,
 };
 
 use crate::{
@@ -120,7 +123,7 @@ impl Default for ListOptions {
 
 impl ListOptions {
     const OPTS: [Opt<ListOptions>; 15] = [
-        opt_todo! {
+        opt_nop! {
             "show-photos",
             |o, s, _| Ok({ o.photos = s; }),
             "display photo IDs during key listings",
@@ -180,7 +183,7 @@ impl ListOptions {
             "show revoked and expired user IDs in key listings",
         },
 
-        opt_todo! {
+        opt! {
             "show-unusable-subkeys",
             |o, s, _| Ok({ o.unusable_subkeys = s; }),
             "show revoked and expired subkeys in key listings",
@@ -230,7 +233,8 @@ impl ListOptions {
 pub fn cmd_list_keys(config: &crate::Config, args: &[String], list_secret: bool)
                      -> Result<()>
 {
-    let mut sink = io::stdout(); // XXX
+    // Key lists are always written to stdout.
+    let mut sink = crate::io::stdout();
 
     // First, emit a header on --list-keys --with-colons.
     if config.with_colons && ! list_secret {
@@ -252,7 +256,7 @@ pub fn cmd_list_keys(config: &crate::Config, args: &[String], list_secret: bool)
         // mess up the tests, it is also likely surprising and
         // confusing for users.
         let trust_root_fp =
-            if let Ok(overlay) = config.keydb().get_certd_overlay()
+            if let Ok(overlay) = config.keydb().certd_overlay()
         {
             if let Ok(trust_root) = overlay.trust_root() {
                 // As prior versions of the Chameleon generated a
@@ -353,7 +357,8 @@ pub fn cmd_list_keys(config: &crate::Config, args: &[String], list_secret: bool)
 pub fn cmd_show_keys(config: &crate::Config, args: &[String])
                      -> Result<()>
 {
-    let sink = io::stdout(); // XXX
+    // Key lists are always written to stdout.
+    let sink = crate::io::stdout();
 
     let mut certs = Vec::new();
 
@@ -439,14 +444,13 @@ where
         config, Some(now), list_all && ! list_secret_keys_mode)?;
     let p = vtm.policy();
 
-    let mut secrets = KeyInfoList::empty();
-    if list_secret || (config.with_secret && config.with_colons) {
-        if let Ok(mut agent) = config.connect_agent().await {
-            if let Ok(s) = agent.list_keys().await {
-                secrets = s;
-            }
-        }
-    }
+    let mut secrets =
+        if list_secret || (config.with_secret && config.with_colons) {
+            KeyInfoProxy::connect(&config).await
+                .unwrap_or_else(|_| KeyInfoProxy::empty())
+        } else {
+            KeyInfoProxy::empty()
+        };
 
     // We emit the location header for humans only if we actually list
     // at least one key.
@@ -456,16 +460,12 @@ where
     let mut sig_stats = SigStats::default();
 
     for cert in certs {
-        let mut has_secret: BTreeSet<Fingerprint> = cert
-            .keys()
-            .filter_map(|k| {
-                if secrets.lookup_by_key(&k).is_some() {
-                    Some(k.fingerprint())
-                } else {
-                    None
-                }
-            })
-            .collect();
+        let mut has_secret = BTreeSet::<Fingerprint>::default();
+        for k in cert.keys() {
+            if secrets.lookup_by_key(&k).await.is_some() {
+                has_secret.insert(k.fingerprint());
+            }
+        }
 
         // When we are importing secret keys, we may have the secret
         // while the agent does not yet have it.  Nevertheless, we
@@ -492,8 +492,13 @@ where
         if list_header && ! emitted_header && ! config.with_colons {
             emitted_header = true;
 
+            // XXX: This is just wrong, or just happens to work, or we
+            // don't properly test for it...  Are we absolutely sure
+            // that we only end up here when there is a certd overlay,
+            // i.e. `--no-default-keyring` was not used?  Maybe...
             let path =
-                config.keydb().get_certd_overlay()?.path().display().to_string();
+                config.keydb().metadata_overlay()?.base_dir()
+                .display().to_string();
             writeln!(&mut sink, "{}", path)?;
             sink.write_all(crate::utils::undeline_for(&path))?;
             writeln!(&mut sink)?;
@@ -505,7 +510,7 @@ where
         let have_secret = has_secret.contains(&cert_fp);
         let ownertrust = config.trustdb.get_ownertrust(&cert_fp)
             .unwrap_or_else(|| OwnerTrustLevel::Undefined.into());
-        let token_sn = secrets.lookup_by_key(cert.primary_key().key())
+        let token_sn = secrets.lookup_by_key(cert.primary_key().key()).await
             .and_then(|i| if let Some(sn) = i.serialno() {
                 Some(TokenSN::SerialNumber(sn.into()))
             } else if have_secret {
@@ -690,14 +695,16 @@ where
 
         for (validity, subkey) in acert.subkeys() {
             // Don't display expired or revoked subkeys.
-            if ! config.with_colons && (validity.expired || validity.revoked) {
+            if ! config.with_colons && (validity.expired || validity.revoked)
+                && ! config.list_options.unusable_subkeys
+            {
                 continue;
             }
 
             let vsubkey = subkey.clone().with_policy(p, now).ok();
             let subkey_fp = subkey.fingerprint();
             let have_secret = has_secret.contains(&subkey_fp);
-            let token_sn = secrets.lookup_by_key(subkey.key())
+            let token_sn = secrets.lookup_by_key(subkey.key()).await
                 .and_then(|i| if let Some(sn) = i.serialno() {
                     Some(TokenSN::SerialNumber(sn.into()))
                 } else if have_secret {
@@ -932,3 +939,101 @@ impl SigStats {
         }
     }
 }
+
+/// Similar to [`KeyInfoList`], but works over restricted agent
+/// connections.
+///
+/// Restricted connections do not allow the `KEYINFO --list` command,
+/// so we transparently fall back to polling.
+pub struct KeyInfoProxy {
+    /// The agent connection for polling if the key list didn't
+    /// succeed.
+    agent: Option<sequoia_gpg_agent::Agent>,
+
+    /// List infos, either polled or initially retrieved via the key
+    /// list.
+    infos: HashMap<Keygrip, KeyInfo>,
+
+    /// Whether we successfully got the list of keys.
+    queried_list: bool,
+
+    /// Whether we have polled the info for this key.
+    queried: HashSet<Keygrip>,
+}
+
+impl KeyInfoProxy {
+    /// Returns an empty set of key infos.
+    ///
+    /// Lookups will always fail.
+    pub fn empty() -> Self {
+        KeyInfoProxy {
+            agent: None,
+            infos: Default::default(),
+            queried_list: false,
+            queried: Default::default(),
+        }
+    }
+
+    /// Connects to the agent, retrieves the key list, falling back to
+    /// polling.
+    pub async fn connect(config: &crate::Config<'_>) -> Result<Self> {
+        let mut secrets = Self::empty();
+
+        let mut a = config.connect_agent().await?;
+        if let Ok(s) = a.list_keys().await {
+            secrets.infos = s.iter()
+                .map(|i| (i.keygrip().clone(), i.clone()))
+                .collect();
+
+            // Since GnuPG 2.3.8, the agent returns an empty list in
+            // restricted mode, whereas previously it returned an
+            // error.
+            secrets.queried_list = ! secrets.infos.is_empty();
+        }
+        secrets.agent = Some(a);
+
+        Ok(secrets)
+    }
+
+    /// Returns a reference to the agent, if connected.
+    pub fn agent(&self) -> Option<&sequoia_gpg_agent::Agent> {
+        self.agent.as_ref()
+    }
+
+    /// Returns a mutable reference to the agent, if connected.
+    pub fn agent_mut(&mut self) -> Option<&mut sequoia_gpg_agent::Agent> {
+        self.agent.as_mut()
+    }
+
+    /// Like [`KeyInfoList::lookup_by_key`], but also polls the
+    /// agent if we couldn't retrieve the full list.
+    pub async fn lookup_by_key<K, R, P>(&mut self, key: K) -> Option<&KeyInfo>
+    where
+        K: Borrow<Key<P, R>>,
+        P: key:: KeyParts,
+        R: key:: KeyRole,
+    {
+        let key = key.borrow();
+        // The mapping from Key to Keygrip is not total so this might
+        // fail.  But if there isn't a keygrip for a key, then
+        // gpg-agent can't manage it, and thus can't return it.
+        // Therefore if a key doesn't have a keygrip, the correct
+        // answer is not to return an error, but to return `None`.
+        let keygrip = Keygrip::of(key.mpis()).ok()?;
+
+        // If it is not in the cache, try polling.
+        if ! self.queried_list && ! self.infos.contains_key(&keygrip) {
+            if let Some(a) = &mut self.agent {
+                if ! self.queried.contains(&keygrip) {
+                    if let Ok(i) = a.key_info(&keygrip).await {
+                        self.infos.insert(keygrip.clone(), i);
+                    }
+
+                    self.queried.insert(keygrip.clone());
+                }
+            }
+        }
+
+        self.infos.get(&keygrip)
+    }
+}
