@@ -103,7 +103,7 @@ const THIRD_PARTY_SIG_THRESHOLD: usize = 250;
 
 /// Synchronize using this file.
 fn lock_path(config: &crate::Config) -> io::Result<PathBuf> {
-    let p = config.keydb().get_certd_overlay()
+    let p = config.keydb().certd_overlay()
         .map_err(|e| io::Error::new(io::ErrorKind::NotFound, e))?
         .path();
     Ok(p.join("_sequoia_parcimonie"))
@@ -202,7 +202,7 @@ fn real_start(config: &crate::Config) -> Result<()> {
         .arg("--homedir").arg(&config.homedir);
 
     for ks in &config.keyserver {
-        p.arg("--keyserver").arg(ks.url());
+        p.arg("--keyserver").arg(ks.url().to_string());
     }
 
     if let Some(modules) = crate::tracing::enabled_modules() {
@@ -449,7 +449,7 @@ async fn worker(config: &mut crate::Config<'_>) -> openpgp::Result<()> {
                     .clone()
                     .for_url(k.url())?
                     .build()?;
-	        net::KeyServer::with_client(k.url(), c)
+	        net::KeyServer::with_client(&k.url().to_string(), c)
 	    }).collect::<Result<Vec<_>>>()?
         } else {
             t!("No keyserver access allowed.");
