@@ -50,11 +50,7 @@ pub fn cmd_quick_add_uid(config: &mut crate::Config, args: &[String])
 async fn add_uid(config: &mut crate::Config<'_>, query: Query, new_uid: UserID)
                  -> Result<()>
 {
-    // Lookup without groups.
-    let certs = config.lookup_certs_with(
-        config.trust_model_impl.with_policy(config, Some(config.now()))?.as_ref(),
-        &query,
-        false)?;
+    let certs = config.lookup_certs(&query)?;
 
     // Cowardly refuse any queries that resolve to multiple keys.  In
     // my mind, using queries other than fingerprints is fragile and
@@ -151,11 +147,7 @@ pub fn cmd_quick_revoke_uid(config: &mut crate::Config, args: &[String])
 async fn revoke_uid(config: &mut crate::Config<'_>, query: Query, rev_uid: UserID)
                     -> Result<()>
 {
-    // Lookup without groups.
-    let certs = config.lookup_certs_with(
-        config.trust_model_impl.with_policy(config, Some(config.now()))?.as_ref(),
-        &query,
-        false)?;
+    let certs = config.lookup_certs(&query)?;
 
     // Cowardly refuse any queries that resolve to multiple keys.  In
     // my mind, using queries other than fingerprints is fragile and
