@@ -6,7 +6,7 @@ use std::{
     fmt,
     io,
     io::Write,
-    sync::Mutex,
+    sync::{Mutex, atomic::{AtomicBool, Ordering}},
     time::SystemTime,
 };
 
@@ -40,6 +40,9 @@ use crate::{
 pub struct Fd {
     stream: Option<Mutex<RefCell<Box<dyn io::Write + Send + Sync>>>>,
     exit_on_write_error: bool,
+
+    /// Did we emit a failure?
+    failure_emitted: AtomicBool,
 }
 
 impl<S: io::Write + Send + Sync + 'static> From<S> for Fd {
@@ -47,6 +50,7 @@ impl<S: io::Write + Send + Sync + 'static> From<S> for Fd {
         Fd {
             stream: Some(Mutex::new(RefCell::new(Box::new(s)))),
             exit_on_write_error: false,
+            failure_emitted: false.into(),
         }
     }
 }
@@ -57,6 +61,7 @@ impl Fd {
         Fd {
             stream: None,
             exit_on_write_error: false,
+            failure_emitted: false.into(),
         }
     }
 
@@ -81,7 +86,26 @@ impl Fd {
     #[allow(dead_code)]
     pub fn emit(&self, status: Status<'_>) -> Result<()> {
         crate::with_invocation_log(|sink| status.emit(sink));
+
         if let Some(fd) = self.stream.as_ref() {
+            match status {
+                Status::Failure { .. } =>
+                    if self.failure_emitted.fetch_or(true, Ordering::SeqCst)
+                {
+                    // We previously emitted a failure.  GnuPG only
+                    // emits the first encountered failure.
+                    return Ok(());
+                },
+
+                Status::BadSig { .. }
+                | Status::ErrSig { .. }
+                | Status::DecryptionFailed { .. } =>
+                    self.failure_emitted.store(true, Ordering::SeqCst),
+
+                _ => (),
+            };
+
+
             let r = status.emit(
                 &mut *fd.lock().expect("not poisoned").borrow_mut());
 
@@ -93,6 +117,11 @@ impl Fd {
         Ok(())
     }
 
+    /// Suppresses the implicit failure message.
+    pub fn suppress_implicit_failure(&self) {
+        self.failure_emitted.store(true, Ordering::SeqCst);
+    }
+
     /// Emits a status message or an interactive prompt.
     ///
     /// If --status-fd was passed, then emits something on the
@@ -103,12 +132,36 @@ impl Fd {
         if self.enabled() {
             self.emit(status)
         } else {
-            let mut sink = io::stdout();
+            let mut sink = crate::io::tty_out();
             write!(&mut sink, "{} ", msg)?;
             sink.flush()?;
             Ok(())
         }
     }
+
+    /// Emits a failure message.
+    ///
+    /// Note: if an error has been already emitted, this does nothing.
+    /// Only the first failure message will be emitted, as in GnuPG.
+    pub fn failure(&self, location: &str, error: crate::error_codes::Error)
+                   -> Result<()>
+    {
+        self.emit(Status::Failure { location, error })
+    }
+
+    /// Emits a failure message.
+    ///
+    /// Note: if an error has been already emitted, this does nothing.
+    /// Only the first failure message will be emitted, as in GnuPG.
+    pub fn failure_from_error(&self, location: &str, _error: &anyhow::Error)
+                              -> Result<()>
+    {
+        // XXX: Map errors to codes.
+        self.emit(Status::Failure {
+            location,
+            error: crate::error_codes::Error::GPG_ERR_GENERAL,
+        })
+    }
 }
 
 #[allow(dead_code)]
@@ -255,6 +308,8 @@ pub enum Status<'a> {
         at: SystemTime,
     },
 
+    KeyRevoked,
+
     NoPubkey {
         issuer: KeyID,
     },
@@ -676,6 +731,10 @@ impl Status<'_> {
                 writeln!(w, "KEYEXPIRED {}", t.format("%s"))?;
             },
 
+            KeyRevoked => {
+                writeln!(w, "KEYREVOKED")?;
+            },
+
             NoPubkey {
                 issuer,
             } => {
