@@ -63,6 +63,7 @@ pub fn open_multiple(control: &dyn common::Common, names: &[String])
                              "can't open signed data '{}'", name));
                          control.error(format_args!(
                              "can't hash datafile: {}", e));
+                         control.status().suppress_implicit_failure();
                          e
                      })?)
         });
@@ -107,15 +108,42 @@ impl io::Read for MultiReader {
 /// Opens a (special) file for writing.
 pub fn create(control: &dyn common::Common, name: &str)
               -> Result<Box<dyn io::Write + Send + Sync>>
+{
+    create_intern(control, name, None)
+}
+
+/// Opens a (special) file for writing, adding an extension to `name`
+/// unless `name` refers to a file descriptor.
+pub fn create_with_extension(control: &dyn common::Common, name: &str,
+                             add_extension: &str)
+                             -> Result<Box<dyn io::Write + Send + Sync>>
+{
+    create_intern(control, name, Some(add_extension))
+}
+
+/// Opens a (special) file for writing, optionally adding an extension
+/// to `name` unless `name` refers to a file descriptor.
+fn create_intern(control: &dyn common::Common,
+                 name: &str,
+                 add_extension: Option<&str>)
+                 -> Result<Box<dyn io::Write + Send + Sync>>
 {
     if name == "-" {
-        Ok(Box::new(io::stdout()))
+        Ok(Box::new(crate::io::stdout()))
     } else if control.special_filenames()
         && special_filename_fd(name).is_some()
     {
         let fd = special_filename_fd(name).expect("checked above");
         sink_from_fd(fd)
     } else {
+        let name_;
+        let mut name = name;
+
+        if let Some(e) = add_extension {
+            name_ = format!("{}.{}", name, e);
+            name = &name_;
+        }
+
         Ok(Box::new(fs::File::create(name)?))
     }
 }
