@@ -13,6 +13,7 @@ use openpgp::{
     packet::{
         Packet,
         Signature,
+        Tag,
         header::BodyLength,
     },
     packet::signature::subpacket::*,
@@ -91,7 +92,7 @@ impl Default for VerifyOptions {
 
 impl VerifyOptions {
     const OPTS: [Opt<VerifyOptions>; 12] = [
-        opt_todo! {
+        opt_nop! {
             "show-photos",
             |o, s, _| Ok({ o.photos = s; }),
             "display photo IDs during signature verification",
@@ -103,25 +104,25 @@ impl VerifyOptions {
             "show policy URLs during signature listings",
         },
 
-        opt_todo! {
+        opt! {
             "show-notations",
             |o, s, _| Ok({ o.ietf_notations = s; o.user_notations = s; }),
             "show all notations during signature listings",
         },
 
-        opt_todo! {
+        opt! {
             "show-std-notations",
             |o, s, _| Ok({ o.ietf_notations = s; }),
             "show IETF standard notations during signature listings",
         },
 
-        opt_todo! {
+        opt! {
             "show-standard-notations",
             |o, s, _| Ok({ o.ietf_notations = s; }),
             "",
         },
 
-        opt_todo! {
+        opt! {
             "show-user-notations",
             |o, s, _| Ok({ o.user_notations = s; }),
             "show user-supplied notations during signature listings",
@@ -216,6 +217,9 @@ pub fn cmd_verify(control: &crate::Config, args: &[String])
 
     let policy = control.policy();
 
+    // The encountered packets, represented by their tags.
+    let mut packets = Vec::with_capacity(16);
+
     // Long story short: If we have at least two files, it is a
     // detached signature.
     let do_it = || -> Result<()> {
@@ -227,12 +231,14 @@ pub fn cmd_verify(control: &crate::Config, args: &[String])
             v.verify_reader(data)?;
             Ok(())
         } else {
-            let mut sink = if let Some(name) = control.outfile() {
-                utils::create(control, name)?
-            } else {
-                Box::new(io::sink())
-            };
-            let helper = VHelper::new(control, 1);
+            let mut sink = control.sink()?;
+
+            // We record a trace of the packets, because we want to
+            // detect the case of trying to verify a detached
+            // signature without specifying the data file.  We will
+            // use this in map_verificaton_error to emit a specific
+            // error message.
+            let helper = VHelper::record_packets(control, &mut packets, 1);
             let mut v = VerifierBuilder::from_reader(sig)?
                 .with_policy(policy, control.now(), helper)?;
             io::copy(&mut v, &mut sink)?;
@@ -240,7 +246,7 @@ pub fn cmd_verify(control: &crate::Config, args: &[String])
         }
     };
 
-    map_verificaton_error(control, do_it())
+    map_verificaton_error(control, do_it(), &packets)
 }
 
 /// Dispatches the --verify-files command.
@@ -271,11 +277,7 @@ pub fn cmd_verify_files(control: &crate::Config, args: &[String])
             // Curiously, GnuPG supports --output with --multifile
             // --verify, but it will override the file with each
             // verification.
-            let mut sink = if let Some(name) = control.outfile() {
-                utils::create(control, name)?
-            } else {
-                Box::new(io::sink())
-            };
+            let mut sink = control.sink()?;
             let helper = VHelper::new(control, 1);
             let mut v = VerifierBuilder::from_reader(sig)?
                 .with_policy(policy, control.now(), helper)?;
@@ -284,48 +286,84 @@ pub fn cmd_verify_files(control: &crate::Config, args: &[String])
             Ok(())
         };
 
-        let _ = map_verificaton_error(control, do_it());
+        let _ = map_verificaton_error(control, do_it(), &[]);
         control.status().emit(Status::FileDone)?;
     }
 
     Ok(())
 }
 
-fn map_verificaton_error(control: &crate::Config, r: Result<()>)
+fn map_verificaton_error(control: &crate::Config, r: Result<()>, packets: &[Tag])
                          -> Result<()> {
     match r {
         Ok(()) => Ok(()),
         Err(e) => {
-            control.override_status_code(1);
-            match e.downcast::<openpgp::Error>() {
-                Ok(oe) => {
-                    // Map our errors to the way GnuPG reports errors.
-                    match oe {
-                        openpgp::Error::MalformedPacket(_) =>
-                        {
-                            control.status().emit(Status::NoData(
-                                NoDataReason::ExpectedPacket))?;
-                            control.status().emit(Status::NoData(
-                                NoDataReason::ExpectedSignature))?;
-                        },
-                        openpgp::Error::MalformedMessage(_) => {
-                            control.status().emit(Status::NoData(
-                                NoDataReason::InvalidPacket))?;
-                            control.status().emit(Status::NoData(
-                                NoDataReason::ExpectedSignature))?;
-                        },
-                        _ => (),
-                    }
-                    Err(oe.into())
-                },
-                Err(e) => Err(e),
+            // If we were verifying inline-signed messages, see if the
+            // user tried to verify a detached signature instead, but
+            // forgot to specify the data file.
+            //
+            // If that happens, we assumed above that we are verifying
+            // an inline-signed message, but only encountered
+            // signature packets, hence the verification failed.
+            if ! packets.is_empty()
+                && packets.iter().all(|t| t == &Tag::Signature)
+            {
+                control.error(format_args!(
+                    "no signed data"));
+                control.error(format_args!(
+                    "can't hash datafile: No data"));
+                control.override_status_code(2);
+                control.status().suppress_implicit_failure();
+                return Err(e);
             }
+
+            let emit_nodata = || -> Result<()> {
+                control.status().emit(Status::NoData(
+                    NoDataReason::ExpectedPacket))?;
+                control.status().emit(Status::NoData(
+                    NoDataReason::ExpectedSignature))?;
+                Ok(())
+            };
+
+            if let Some(oe) = e.downcast_ref::<openpgp::Error>() {
+                // Map our errors to the way GnuPG reports errors.
+                match oe {
+                    openpgp::Error::MalformedPacket(_) =>
+                        emit_nodata()?,
+                    openpgp::Error::MalformedMessage(_) =>
+                        emit_nodata()?,
+                    _ => (),
+                }
+            }
+
+            if let Some(ioe) = e.downcast_ref::<io::Error>() {
+                match ioe.kind() {
+                    io::ErrorKind::UnexpectedEof => {
+                        control.error(format_args!(
+                            "no valid OpenPGP data found."));
+                        emit_nodata()?;
+	                control.error(format_args!(
+                            "the signature could not be verified.\n\
+                             Please remember that the signature file (.sig or .asc)\n\
+		             should be the first file given on the command line.\n"));
+                        control.override_status_code(2);
+                    },
+                    _ => (),
+                }
+            }
+
+            control.override_status_code(1);
+            Err(e)
         },
     }
 }
 
 pub struct VHelper<'a, 'store> {
     control: &'a crate::Config<'store>,
+
+    /// Optionally we record a packet trace.
+    packets: Option<&'a mut Vec<Tag>>,
+
     #[allow(dead_code)]
     signatures: usize,
     good_signatures: usize,
@@ -347,6 +385,26 @@ impl<'a, 'store> VHelper<'a, 'store> {
                -> Self {
         VHelper {
             control,
+            packets: None,
+            signatures,
+            good_signatures: 0,
+            good_checksums: 0,
+            unknown_checksums: 0,
+            bad_signatures: 0,
+            bad_checksums: 0,
+            broken_signatures: 0,
+            weak_digest_warning_printed: Default::default(),
+        }
+    }
+
+    pub fn record_packets(control: &'a crate::Config<'store>,
+                          packets: &'a mut Vec<Tag>,
+                          signatures: usize)
+                         -> Self
+    {
+        VHelper {
+            control,
+            packets: Some(packets),
             signatures,
             good_signatures: 0,
             good_checksums: 0,
@@ -611,13 +669,21 @@ impl<'a, 'store> VHelper<'a, 'store> {
             })?;
         }
 
-        // First, GnuPG emits a key considered status as a side-effect
+        // First, gpgv emits a key considered status as a side-effect
         // of evaluating the trust information.  Emulate that.
-        self.control.status().emit(Status::KeyConsidered {
-            fingerprint: ka.cert().fingerprint(),
-            not_selected: false,
-            all_expired_or_revoked: false // XXX: I haven't seen GnuPG set that.
-        })?;
+        if self.control.argv0() == "gpgv" {
+            self.control.status().emit(Status::KeyConsidered {
+                fingerprint: ka.cert().fingerprint(),
+                not_selected: false,
+                all_expired_or_revoked: false // XXX: I haven't seen GnuPG set that.
+            })?;
+        }
+
+        let validity =
+            self.control.lookup_certs(
+                &Query::ExactKey(ka.cert().key_handle()))?
+            .get(0)
+            .map(|(validity, _cert)| *validity);
 
         let primary_uid = crate::utils::best_effort_primary_uid(
             self.control.policy(), ka.cert());
@@ -647,43 +713,6 @@ impl<'a, 'store> VHelper<'a, 'store> {
             e => unimplemented!("{:?}", e),
         }
 
-        // Dump notations.
-        for notation in sig.notation_data() {
-            self.control.status().emit(Status::NotationName {
-                name: notation.name().into(),
-            })?;
-            if notation.flags().human_readable() {
-                self.control.status().emit(Status::NotationFlags {
-                    // If it were critical, the sig would
-                    // not have checked out
-                    critical: false,
-                    human_readable: true,
-                })?;
-            }
-            self.control.status().emit(Status::NotationData {
-                data: notation.value().into(),
-            })?;
-        }
-
-        // Cryptographically valid.
-        self.control.status().emit(Status::ValidSig {
-            issuer: ka.fingerprint(),
-            creation_time: sig.signature_creation_time()
-                .expect("every well-formed signature has one"),
-            expire_time: sig.signature_expiration_time(),
-            version: sig.version(),
-            pk_algo: sig.pk_algo(),
-            hash_algo: sig.hash_algo(),
-            sig_class: sig.typ(),
-            primary: ka.cert().fingerprint(),
-        })?;
-
-        let validity =
-            self.control.lookup_certs(
-                &Query::ExactKey(ka.cert().key_handle()))?
-            .get(0)
-            .map(|(validity, _cert)| *validity);
-
         if let Some(v) = validity.as_ref()
             .filter(|_| self.control.verify_options.uid_validity)
         {
@@ -701,6 +730,154 @@ impl<'a, 'store> VHelper<'a, 'store> {
             }
         }
 
+        match error.filter(|_| ! crate::gnupg_interface::STRICT_OUTPUT
+                           || self.control.argv0() != "gpgv")
+        {
+            Some(openpgp::Error::InvalidKey(_)) => {
+                // This is fun!  In GnuPG, the cert revocation takes
+                // precedence over the subkey revocation for the
+                // warning text, but when displaying the reason for
+                // revocation, it prefers the subkey revocation.
+
+                // First, the warning text, preferring the cert
+                // revocation.
+                if let RevocationStatus::Revoked(_) =
+                    ka.cert().revocation_status()
+                {
+                    self.control.warn(format_args!(
+                        "WARNING: This key has been revoked by its owner!"));
+                    self.control.warn(format_args!(
+                        "         This could mean that the signature is forged."));
+                } else if let RevocationStatus::Revoked(_) =
+                    ka.revocation_status()
+                {
+                    self.control.warn(format_args!(
+                        "WARNING: This subkey has been revoked by its owner!"));
+                }
+
+                // Then, display information about the revocation,
+                // preferring the subkey revocation.
+                if let RevocationStatus::Revoked(rev) =
+                    ka.revocation_status()
+                {
+                    for sig in rev {
+                        if let Some((rfr, msg)) = sig.reason_for_revocation() {
+                            self.control.warn(format_args!(
+                                "reason for revocation: {}", babel::Fish(rfr)));
+                            self.control.warn(format_args!(
+                                "revocation comment: {}",
+                                String::from_utf8_lossy(msg)));
+                        }
+                    }
+                } else if let RevocationStatus::Revoked(rev) =
+                    ka.cert().revocation_status()
+                {
+                    for sig in rev {
+                        if let Some((rfr, msg)) = sig.reason_for_revocation() {
+                            self.control.warn(format_args!(
+                                "reason for revocation: {}", babel::Fish(rfr)));
+                            self.control.warn(format_args!(
+                                "revocation comment: {}",
+                                String::from_utf8_lossy(msg)));
+                        }
+                    }
+                }
+            },
+            _ => (),
+        }
+
+        // Dump notations.
+        for notation in sig.notation_data()
+            .collect::<Vec<_>>().into_iter().rev()
+        {
+            self.control.status().emit(Status::NotationName {
+                name: notation.name().into(),
+            })?;
+            if notation.flags().human_readable() {
+                self.control.status().emit(Status::NotationFlags {
+                    // If it were critical, the sig would
+                    // not have checked out
+                    critical: false,
+                    human_readable: true,
+                })?;
+            }
+            self.control.status().emit(Status::NotationData {
+                data: notation.value().into(),
+            })?;
+
+            // Shall we display it to the user?
+            let user_notation = notation.name().contains("@");
+            if (user_notation && self.control.verify_options.user_notations)
+                || (! user_notation && self.control.verify_options.ietf_notations)
+            {
+                self.control.warn(format_args!(
+                    "Signature notation: {}={}",
+                    notation.name(),
+                    format_notation_value(notation)));
+            }
+        }
+
+        // Cryptographically valid.
+        self.control.status().emit(Status::ValidSig {
+            issuer: ka.fingerprint(),
+            creation_time: sig.signature_creation_time()
+                .expect("every well-formed signature has one"),
+            expire_time: sig.signature_expiration_time(),
+            version: sig.version(),
+            pk_algo: sig.pk_algo(),
+            hash_algo: sig.hash_algo(),
+            sig_class: sig.typ(),
+            primary: ka.cert().fingerprint(),
+        })?;
+
+        // Consider the key some more, for good measure.
+        if self.control.argv0() == "gpg" {
+            for _ in 0..2 {
+                self.control.status().emit(Status::KeyConsidered {
+                    fingerprint: ka.cert().fingerprint(),
+                    not_selected: false,
+                    all_expired_or_revoked: false,
+                })?;
+            }
+        }
+
+        match error.filter(|_| ! crate::gnupg_interface::STRICT_OUTPUT
+                           || self.control.argv0() != "gpgv")
+        {
+            Some(openpgp::Error::InvalidKey(_)) => {
+                if let RevocationStatus::Revoked(_) =
+                    ka.revocation_status()
+                {
+                    self.control.status().emit(Status::KeyRevoked)?;
+
+                    // And consider the key some more, being
+                    // considerate never hurt anyone.
+                    self.control.status().emit(Status::KeyConsidered {
+                        fingerprint: ka.cert().fingerprint(),
+                        not_selected: false,
+                        all_expired_or_revoked: false,
+                    })?;
+                } else if let RevocationStatus::Revoked(_) =
+                    ka.cert().revocation_status()
+                {
+                    self.control.status().emit(Status::KeyRevoked)?;
+
+                    // And consider the key some more, extra
+                    // carefully.  No-one was ever fired for being too
+                    // considerate.
+                    for _ in 0..2 {
+                        self.control.status().emit(Status::KeyConsidered {
+                            fingerprint: ka.cert().fingerprint(),
+                            not_selected: false,
+                            all_expired_or_revoked: false,
+                        })?;
+                    }
+                }
+            }
+
+            _ => (),
+        }
+
         let print_fingerprint = if validity.map(|v| v.revoked).unwrap_or(false) {
             // XXX
             false
@@ -954,6 +1131,11 @@ impl<'a, 'store> VHelper<'a, 'store> {
 
 impl<'a, 'store> VerificationHelper for VHelper<'a, 'store> {
     fn inspect(&mut self, pp: &openpgp::parse::PacketParser) -> Result<()> {
+        // Record a packet trace if enabled.
+        if let Some(d) = self.packets.as_mut() {
+            d.push(pp.packet.tag());
+        }
+
         match &pp.packet {
             Packet::Literal(p) if ! self.control.list_only => {
                 self.control.status().emit(
@@ -1039,3 +1221,31 @@ fn key_flags_to_usage(f: KeyFlags) -> u8 {
         .clear_split_key()
         .is_empty()                   { GCRY_PK_USAGE_UNKN } else { 0 }
 }
+
+/// Formats the notation data, escaping and truncating
+/// non-human-readable data.
+pub fn format_notation_value(n: &NotationData) -> String {
+    if n.flags().human_readable() {
+        String::from_utf8_lossy(n.value())
+            .to_string()
+    } else {
+        let preview: String =
+            n.value().iter().take(19)
+            .map(|b|
+                 char::from_u32((*b).into())
+                 .filter(|c| *c == ' '
+                         || c.is_ascii_graphic())
+                 .unwrap_or('?'))
+            .collect();
+
+        format!("[ not human readable \
+                 ({} bytes: {}{}) ]",
+                n.value().len(),
+                preview,
+                if n.value().len() > 20 {
+                    "..."
+                } else {
+                    ""
+                })
+    }
+}
