@@ -341,7 +341,19 @@ fn disabled() -> Result<()> {
     diff.assert_success();
     diff.assert_limits(0, 150, 67);
 
-    test_key_cert_imported(cert, experiment)
+    test_key_cert_imported(&mut experiment)?;
+
+    experiment.section("Importing TSK...");
+    let diff = experiment.invoke(&[
+        "--batch",
+        "--import",
+        &experiment.store("cert", &cert.as_tsk().to_vec()?)?,
+    ])?;
+    diff.assert_success();
+
+    test_key_key_imported(&mut experiment)?;
+
+    Ok(())
 }
 
 #[test]
@@ -415,6 +427,46 @@ mhZeqo6zyn8zrO9RGU7+8jmeb5nVnXw1YmZcw2fiJgI9+tTMkTfomyR6k0EDvcEu
     test_key(cert, experiment)
 }
 
+#[test]
+#[ntest::timeout(600000)]
+fn restricted_agent() -> Result<()> {
+    let mut experiment = make_experiment!()?;
+    let cert = experiment.artifact(
+        "cert",
+        ||  CertBuilder::new()
+            .set_creation_time(Experiment::now())
+            .add_userid("Alice Lovelace <alice@lovelace.name>")
+            .add_signing_subkey()
+            .generate()
+            .map(|(cert, _rev)| cert),
+        |a, f| a.as_tsk().serialize(f),
+        |b| Cert::from_bytes(&b))?;
+
+    experiment.section("Importing cert...");
+    let diff = experiment.invoke(&[
+        "--import",
+        &experiment.store("cert", &cert.to_vec()?)?,
+    ])?;
+    diff.assert_success();
+
+    experiment.restrict_agent()?;
+    test_key_cert_imported(&mut experiment)?;
+
+    experiment.unrestrict_agent()?;
+    experiment.section("Importing TSK...");
+    let diff = experiment.invoke(&[
+        "--batch",
+        "--import",
+        &experiment.store("cert", &cert.as_tsk().to_vec()?)?,
+    ])?;
+    diff.assert_success();
+
+    experiment.restrict_agent()?;
+    test_key_key_imported(&mut experiment)?;
+
+    Ok(())
+}
+
 fn test_key(cert: Cert, mut experiment: Experiment) -> Result<()>
 {
     experiment.section("Importing cert...");
@@ -424,10 +476,22 @@ fn test_key(cert: Cert, mut experiment: Experiment) -> Result<()>
     ])?;
     diff.assert_success();
 
-    test_key_cert_imported(cert, experiment)
+    test_key_cert_imported(&mut experiment)?;
+
+    experiment.section("Importing TSK...");
+    let diff = experiment.invoke(&[
+        "--batch",
+        "--import",
+        &experiment.store("cert", &cert.as_tsk().to_vec()?)?,
+    ])?;
+    diff.assert_success();
+
+    test_key_key_imported(&mut experiment)?;
+
+    Ok(())
 }
 
-fn test_key_cert_imported(cert: Cert, mut experiment: Experiment) -> Result<()>
+fn test_key_cert_imported(experiment: &mut Experiment) -> Result<()>
 {
     let diff = experiment.invoke(&[
         "--list-keys",
@@ -590,14 +654,11 @@ fn test_key_cert_imported(cert: Cert, mut experiment: Experiment) -> Result<()>
     diff.assert_success();
     diff.assert_equal_up_to(0, 0);
 
-    experiment.section("Importing TSK...");
-    let diff = experiment.invoke(&[
-        "--batch",
-        "--import",
-        &experiment.store("cert", &cert.as_tsk().to_vec()?)?,
-    ])?;
-    diff.assert_success();
+    Ok(())
+}
 
+fn test_key_key_imported(experiment: &mut Experiment) -> Result<()>
+{
     let diff = experiment.invoke(&[
         "--list-keys",
     ])?;
@@ -1241,6 +1302,116 @@ fn unusable_uids() -> Result<()> {
     Ok(())
 }
 
+#[test]
+#[ntest::timeout(600000)]
+fn show_unusable_subkeys() -> Result<()> {
+    let mut experiment = make_experiment!()?;
+
+    let year = Duration::new(365 * 24 * 60 * 60, 0);
+    let t0 = Experiment::now() - 3 * year;
+    let t1 = Experiment::now() - 2 * year;
+    let p = openpgp::policy::StandardPolicy::new();
+
+    // Create a key where one signing subkey is expired, and a newer
+    // one that is not expired, but where the secret isn't available.
+    let cert = experiment.artifact(
+        "cert", || {
+            let (c, _) = CertBuilder::new()
+                .add_userid("Alice Lovelace <alice@lovelace.name>")
+                .set_creation_time(t0)
+                .generate()?;
+
+            // Now add a signing subkey that is expired.
+            let mut subkey: Key<_, key::SubordinateRole> =
+                key::Key4::generate_ecc(true, Curve::Ed25519)?.into();
+            subkey.set_creation_time(t0)?;
+
+            let builder = signature::SignatureBuilder::new(
+                SignatureType::SubkeyBinding)
+                .set_signature_creation_time(t0)?
+                .set_key_expiration_time(&subkey, t1)?
+                .set_key_flags(KeyFlags::empty().set_signing())?;
+            let c = SubkeyBuilder::new_with(c.with_policy(&p, None)?,
+                                            subkey, builder)
+                .attach_cert()?;
+
+            // Now add a signing subkey that is not expired.
+            let mut subkey: Key<_, key::SubordinateRole> =
+                key::Key4::generate_ecc(true, Curve::Ed25519)?.into();
+            subkey.set_creation_time(t1)?;
+
+            let newer_subkey_fp = subkey.fingerprint();
+            let builder = signature::SignatureBuilder::new(
+                SignatureType::SubkeyBinding)
+                .set_signature_creation_time(t1)?
+                .set_key_flags(KeyFlags::empty().set_signing())?;
+            let c = SubkeyBuilder::new_with(c.with_policy(&p, None)?,
+                                            subkey, builder)
+                .attach_cert()?;
+
+            let filtered = c.as_tsk()
+                .set_filter(move |k| k.fingerprint() != newer_subkey_fp)
+                .to_vec()?;
+            Cert::from_bytes(&filtered)
+        },
+        |a, f| a.as_tsk().serialize(f),
+        |b| Cert::from_bytes(&b))?;
+
+    experiment.section("Importing cert...");
+    let diff = experiment.invoke(&[
+        "--import",
+        &experiment.store("cert", &cert.to_vec()?)?,
+    ])?;
+    diff.assert_success();
+
+    let diff = experiment.invoke(&[
+        "--list-keys",
+        "--with-subkey-fingerprint",
+    ])?;
+    diff.assert_success();
+    diff.assert_limits(1, 0, 0);
+
+    let diff = experiment.invoke(&[
+        "--list-keys",
+        "--with-colons",
+    ])?;
+    diff.assert_success();
+    diff.assert_limits(0, 0, 0);
+
+    let diff = experiment.invoke(&[
+        "--list-keys",
+        "--list-options", "no-show-unusable-subkeys",
+    ])?;
+    diff.assert_success();
+    diff.assert_limits(1, 0, 0);
+
+    let diff = experiment.invoke(&[
+        "--list-keys",
+        "--with-colons",
+        "--list-options", "no-show-unusable-subkeys",
+    ])?;
+    diff.assert_success();
+    diff.assert_limits(0, 0, 0);
+
+    let diff = experiment.invoke(&[
+        "--list-keys",
+        "--list-options", "show-unusable-subkeys",
+    ])?;
+    diff.assert_success();
+    diff.assert_limits(1, 0, 0);
+
+    let diff = experiment.invoke(&[
+        "--list-keys",
+        "--with-colons",
+        "--list-options", "show-unusable-subkeys",
+    ])?;
+    diff.assert_success();
+    diff.assert_limits(0, 0, 0);
+
+    Ok(())
+}
+
+
 #[test]
 #[ntest::timeout(600000)]
 fn policy_url() -> Result<()> {
@@ -1563,3 +1734,97 @@ fn show_keys() -> Result<()> {
 
     Ok(())
 }
+
+#[test]
+#[ntest::timeout(600000)]
+fn keyring_usage() -> Result<()> {
+    let mut experiment = make_experiment!()?;
+    let cert = experiment.artifact(
+        "cert",
+        ||  CertBuilder::new()
+            .set_creation_time(Experiment::now())
+            .add_userid("Alice Lovelace")
+            .add_userid("<alice@lovelace.name>")
+            .add_userid("Alice Lovelace <alice@lovelace.name>")
+            .add_signing_subkey()
+            .generate()
+            .map(|(cert, _rev)| cert),
+        |a, f| a.as_tsk().serialize(f),
+        |b| Cert::from_bytes(&b))?;
+
+    // First, use the empty state.
+    t(&mut experiment)?;
+
+    // Then, import the cert, and retry.
+    let diff = experiment.invoke(&[
+        "--import",
+        &experiment.store("cert", &cert.to_vec()?)?,
+    ])?;
+    diff.assert_success();
+
+    // Then, import the key, and retry.
+    let diff = experiment.invoke(&[
+        "--import",
+        &experiment.store("key", &cert.as_tsk().to_vec()?)?,
+    ])?;
+    diff.assert_success();
+
+    fn t(experiment: &mut Experiment) -> Result<()> {
+        for (mode, status_limit) in [
+            ("--no-disable-mdc", 0),
+            ("--no-default-keyring", 0),
+            ("--no-keyring", 37),
+        ] {
+            let diff = experiment.invoke(&[
+                mode,
+                "--list-keys",
+            ])?;
+            diff.assert_success();
+            diff.assert_limits(0, 0, status_limit);
+
+            let diff = experiment.invoke(&[
+                mode,
+                "--list-keys",
+                "--with-colons",
+            ])?;
+            diff.assert_success();
+            diff.assert_limits(3, 0, status_limit);
+
+            let diff = experiment.invoke(&[
+                mode,
+                "--list-keys",
+                "--with-secret",
+            ])?;
+            diff.assert_success();
+            diff.assert_limits(0, 0, status_limit);
+
+            let diff = experiment.invoke(&[
+                mode,
+                "--list-keys",
+                "--with-secret",
+                "--with-colons",
+            ])?;
+            diff.assert_success();
+            diff.assert_limits(3, 0, status_limit);
+
+            let diff = experiment.invoke(&[
+                mode,
+                "--list-secret-keys",
+            ])?;
+            diff.assert_success();
+            diff.assert_limits(0, 0, status_limit);
+
+            let diff = experiment.invoke(&[
+                mode,
+                "--list-secret-keys",
+                "--with-colons",
+            ])?;
+            diff.assert_success();
+            diff.assert_limits(0, 0, status_limit);
+        }
+
+        Ok(())
+    }
+
+    Ok(())
+}
