@@ -7,7 +7,7 @@ use anyhow::Result;
 use sequoia_openpgp as openpgp;
 use openpgp::{
     cert::prelude::*,
-    packet::{*, signature::*, key::*},
+    packet::{*, signature::{*, subpacket::*}, key::*},
     types::*,
     parse::Parse,
     policy::StandardPolicy,
@@ -146,6 +146,7 @@ fn hash_algos() -> Result<()> {
                                i == 0, &cert,
                                sig,
                                vec![],
+                               true,
                                i > 0)?;
     }
 
@@ -203,6 +204,7 @@ fn weak_hash_algos() -> Result<()> {
                                i == 0, &cert,
                                sig,
                                vec!["--weak-digest", &algo.to_string()],
+                               true,
                                false)?;
     }
 
@@ -264,6 +266,7 @@ fn signature_types() -> Result<()> {
                                i == 0, &cert,
                                sig,
                                vec![],
+                               true,
                                i < 2)?;
     }
 
@@ -419,6 +422,7 @@ fn extended() -> Result<()> {
                     test_detached_sig_with(&mut experiment,
                                            i == 0, &cert,
                                            sig.clone(), vec![],
+                                           true,
                                            true)?;
                     i += 1;
                 }
@@ -493,6 +497,7 @@ fn wrong_key() -> Result<()> {
         test_detached_sig_with(&mut experiment,
                                i == 0, &cert,
                                sig, vec![],
+                               true,
                                signing_capable)?;
     }
 
@@ -502,13 +507,15 @@ fn wrong_key() -> Result<()> {
 fn test_detached_sig(experiment: &mut Experiment, cert: &Cert, sig: Packet)
                      -> Result<()>
 {
-    test_detached_sig_with(experiment, true, cert, sig, vec![], true)
+    test_detached_sig_with(experiment, true, cert, sig, vec![], true, true)
 }
 
 fn test_detached_sig_with<'a>(experiment: &mut Experiment,
                               pristine_experiment: bool,
                               cert: &Cert,
-                              sig: Packet, extra_args: Vec<&'a str>,
+                              sig: Packet,
+                              extra_args: Vec<&'a str>,
+                              try_gpgv: bool,
                               expect_success: bool)
                               -> Result<()>
 {
@@ -548,26 +555,28 @@ fn test_detached_sig_with<'a>(experiment: &mut Experiment,
         diff.assert_equal_up_to(0, 20);
 
         // Now try gpgv.
-        let empty_keyring = experiment.store("empty", b"")?;
-        let mut args_good = vec![
-            "gpgv",
-            "--keyring", &empty_keyring,
+        if try_gpgv {
+            let empty_keyring = experiment.store("empty", b"")?;
+            let mut args_good = vec![
+                "gpgv",
+                "--keyring", &empty_keyring,
             ];
-        args_good.extend_from_slice(&extra_args);
-        data_good.iter().for_each(|a| args_good.push(a));
-        let diff = experiment.invoke(&args_good)?;
-        diff.assert_failure();
-        diff.assert_equal_up_to(0, 20);
+            args_good.extend_from_slice(&extra_args);
+            data_good.iter().for_each(|a| args_good.push(a));
+            let diff = experiment.invoke(&args_good)?;
+            diff.assert_failure();
+            diff.assert_equal_up_to(0, 20);
 
-        let mut args_bad = vec![
-            "gpgv",
-            "--keyring", &empty_keyring,
+            let mut args_bad = vec![
+                "gpgv",
+                "--keyring", &empty_keyring,
             ];
-        args_bad.extend_from_slice(&extra_args);
-        data_bad.iter().for_each(|a| args_bad.push(a));
-        let diff = experiment.invoke(&args_bad)?;
-        diff.assert_failure();
-        diff.assert_equal_up_to(0, 20);
+            args_bad.extend_from_slice(&extra_args);
+            data_bad.iter().for_each(|a| args_bad.push(a));
+            let diff = experiment.invoke(&args_bad)?;
+            diff.assert_failure();
+            diff.assert_equal_up_to(0, 20);
+        }
 
         experiment.section("Importing cert...");
         let diff = experiment.invoke(&[
@@ -581,7 +590,7 @@ fn test_detached_sig_with<'a>(experiment: &mut Experiment,
     let diff = experiment.invoke(&args_good)?;
     if expect_success {
         diff.assert_success();
-        diff.assert_limits(0, 10, 67);
+        diff.assert_limits(0, 10, 0);
     } else {
         diff.assert_failure();
         diff.assert_equal_up_to(0, 10);
@@ -596,34 +605,36 @@ fn test_detached_sig_with<'a>(experiment: &mut Experiment,
     }
 
     // Now try gpgv.
-    let cert = &experiment.store("cert", &cert.to_vec()?)?;
-    let mut args_good = vec![
-        "gpgv",
-        "--keyring", &cert,
-    ];
-    args_good.extend_from_slice(&extra_args);
-    data_good.iter().for_each(|a| args_good.push(a));
-    let diff = experiment.invoke(&args_good)?;
-    if expect_success {
-        diff.assert_success();
-        //diff.assert_equal_up_to(0, 10);
-    } else {
-        diff.assert_failure();
-        diff.assert_equal_up_to(0, 10);
-    }
+    if try_gpgv {
+        let cert = &experiment.store("cert", &cert.to_vec()?)?;
+        let mut args_good = vec![
+            "gpgv",
+            "--keyring", &cert,
+        ];
+        args_good.extend_from_slice(&extra_args);
+        data_good.iter().for_each(|a| args_good.push(a));
+        let diff = experiment.invoke(&args_good)?;
+        if expect_success {
+            diff.assert_success();
+            //diff.assert_equal_up_to(0, 10);
+        } else {
+            diff.assert_failure();
+            diff.assert_equal_up_to(0, 10);
+        }
 
-    let mut args_bad = vec![
-        "gpgv",
-        "--keyring", &cert,
-    ];
-    args_bad.extend_from_slice(&extra_args);
-    data_bad.iter().for_each(|a| args_bad.push(a));
-    let diff = experiment.invoke(&args_bad)?;
-    diff.assert_failure();
-    if expect_success {
-        //diff.assert_equal_up_to(200, 10);
-    } else {
-        diff.assert_equal_up_to(0, 10);
+        let mut args_bad = vec![
+            "gpgv",
+            "--keyring", &cert,
+        ];
+        args_bad.extend_from_slice(&extra_args);
+        data_bad.iter().for_each(|a| args_bad.push(a));
+        let diff = experiment.invoke(&args_bad)?;
+        diff.assert_failure();
+        if expect_success {
+            //diff.assert_equal_up_to(200, 10);
+        } else {
+            diff.assert_equal_up_to(0, 10);
+        }
     }
 
     Ok(())
@@ -681,7 +692,7 @@ fn show_uid_validity() -> Result<()> {
             diff.assert_limits(0, 6, 0);
         } else {
             diff.assert_success();
-            diff.assert_limits(0, 6, 67);
+            diff.assert_limits(0, 6, 0);
         }
 
         // Note: There is no machine-readable output (--with-colons)
@@ -698,7 +709,7 @@ fn show_uid_validity() -> Result<()> {
             diff.assert_limits(0, 6, 0);
         } else {
             diff.assert_success();
-            diff.assert_limits(0, 6, 67);
+            diff.assert_limits(0, 6, 0);
         }
 
         let diff = experiment.invoke(&[
@@ -712,7 +723,7 @@ fn show_uid_validity() -> Result<()> {
             diff.assert_limits(0, 6, 0);
         } else {
             diff.assert_success();
-            diff.assert_limits(0, 6, 67);
+            diff.assert_limits(0, 6, 0);
         }
     }
 
@@ -742,7 +753,7 @@ fn show_uid_validity() -> Result<()> {
             &experiment.store("msg", &MSG)?,
         ])?;
         diff.assert_success();
-        diff.assert_limits(0, 6, 67);
+        diff.assert_limits(0, 6, 0);
 
         let diff = experiment.invoke(&[
             "--verify",
@@ -751,8 +762,287 @@ fn show_uid_validity() -> Result<()> {
             &experiment.store("msg", &MSG)?,
         ])?;
         diff.assert_success();
-        diff.assert_limits(0, 6, 67);
+        diff.assert_limits(0, 6, 0);
     }
 
     Ok(())
 }
+
+#[test]
+#[ntest::timeout(600000)]
+fn notations() -> Result<()> {
+    let mut experiment = make_experiment!()?;
+    let cert = experiment.artifact(
+        "cert",
+        || CertBuilder::new()
+            .set_creation_time(Experiment::now())
+            .add_userid("Alice Lovelace <alice@lovelace.name>")
+            .add_signing_subkey()
+            .generate()
+            .map(|(cert, _rev)| cert),
+        |a, f| a.as_tsk().serialize(f),
+        |b| Cert::from_bytes(&b))?;
+
+    let sig = experiment.artifact(
+        "sig",
+        || -> Result<Packet> {
+            let mut subkey_signer =
+                cert.keys().subkeys().secret().next().unwrap()
+                .key().clone().into_keypair()?;
+
+            SignatureBuilder::new(SignatureType::Binary)
+                .set_signature_creation_time(Experiment::now())?
+                .add_notation("human-readable@user.example.org",
+                              "a human-readable value",
+                              NotationDataFlags::empty().set_human_readable(),
+                              false)?
+                .add_notation("machine-readable@user.example.org",
+                              "a machine-readable value",
+                              NotationDataFlags::empty(),
+                              false)?
+                .add_notation("human-readable-fake-ietf",
+                              "a human-readable value",
+                              NotationDataFlags::empty().set_human_readable(),
+                              false)?
+                .add_notation("machine-readable-fake-ietf",
+                              "a machine-readable value",
+                              NotationDataFlags::empty(),
+                              false)?
+                .sign_message(&mut subkey_signer, MSG)
+                .map(Into::into)
+        },
+        |a, f| a.serialize(f),
+        |b| Packet::from_bytes(&b))?;
+
+    test_detached_sig_with(&mut experiment, true, &cert, sig,
+                           vec!["--verify-options=show-notations"],
+                           false,
+                           true)
+}
+
+#[test]
+#[ntest::timeout(600000)]
+fn no_data() -> Result<()> {
+    let mut experiment = make_experiment!()?;
+    let cert = experiment.artifact(
+        "cert",
+        || CertBuilder::new()
+            .set_creation_time(Experiment::now())
+            .add_userid("Alice Lovelace <alice@lovelace.name>")
+            .add_signing_subkey()
+            .generate()
+            .map(|(cert, _rev)| cert),
+        |a, f| a.as_tsk().serialize(f),
+        |b| Cert::from_bytes(&b))?;
+
+    let sig = experiment.artifact(
+        "sig",
+        || -> Result<Packet> {
+            let mut subkey_signer =
+                cert.keys().subkeys().secret().next().unwrap()
+                .key().clone().into_keypair()?;
+
+            SignatureBuilder::new(SignatureType::Binary)
+                .set_signature_creation_time(Experiment::now())?
+                .sign_message(&mut subkey_signer, MSG)
+                .map(Into::into)
+        },
+        |a, f| a.serialize(f),
+        |b| Packet::from_bytes(&b))?;
+
+    // Non-existent signed data.
+    let diff = experiment.invoke(&[
+        "--batch",
+        "--verify",
+        "--no-default-keyring",
+        "--keyring", &experiment.store("cert", &cert.to_vec()?)?,
+        &experiment.store("data.sig", &sig.to_vec()?)?,
+        "doesn't-exist",
+    ])?;
+    diff.assert_failure();
+    diff.assert_limits(0, 13, 0);
+
+    // No signed data.
+    let diff = experiment.invoke(&[
+        "--batch",
+        "--verify",
+        "--no-default-keyring",
+        "--keyring", &experiment.store("cert", &cert.to_vec()?)?,
+        &experiment.store("data.sig", &sig.to_vec()?)?,
+    ])?;
+    diff.assert_failure();
+    diff.assert_limits(0, 0, 0);
+
+    // Zero-sized keyring, detached signature.
+    let diff = experiment.invoke(&[
+        "--batch",
+        "--verify",
+        "--no-default-keyring",
+        "--keyring", &experiment.store("zero-keyring", b"")?,
+        &experiment.store("data.sig", &sig.to_vec()?)?,
+        &experiment.store("msg.txt", MSG)?,
+    ])?;
+    diff.assert_failure();
+    diff.assert_limits(0, 6, 0);
+
+    // Truncated keyring, detached signature.
+    let diff = experiment.invoke(&[
+        "--batch",
+        "--verify",
+        "--no-default-keyring",
+        "--keyring", &experiment.store("trunc-keyring", &cert.to_vec()?[..100])?,
+        &experiment.store("data.sig", &sig.to_vec()?)?,
+        &experiment.store("msg.txt", MSG)?,
+    ])?;
+    diff.assert_failure();
+    // Why the big difference?  We emit a rather helpful error message.
+    diff.assert_limits(0, 150, 0);
+
+    Ok(())
+}
+
+#[test]
+#[ntest::timeout(600000)]
+fn revoked_cert() -> Result<()> {
+    let mut experiment = make_experiment!()?;
+    let cert = experiment.artifact(
+        "cert",
+        || -> Result<Cert> {
+            let (cert, rev) = CertBuilder::new()
+                .set_creation_time(Experiment::now() - Duration::new(3600, 0))
+                .add_userid("Alice Lovelace <alice@lovelace.name>")
+                .add_signing_subkey()
+                .generate()?;
+            Ok(cert.insert_packets2(rev)?.0)
+        },
+        |a, f| a.as_tsk().serialize(f),
+        |b| Cert::from_bytes(&b))?;
+
+    let sig = experiment.artifact(
+        "sig",
+        || -> Result<Packet> {
+            let mut signer = cert.keys().subkeys()
+                .nth(0).unwrap().key().clone()
+                .parts_into_secret()?.into_keypair()?;
+
+            SignatureBuilder::new(SignatureType::Binary)
+                .set_signature_creation_time(Experiment::now())?
+                .sign_message(&mut signer, MSG)
+                .map(Into::into)
+        },
+        |a, f| a.serialize(f),
+        |b| Packet::from_bytes(&b))?;
+
+    test_detached_sig_with(&mut experiment,
+                           true, &cert,
+                           sig, vec![],
+                           true,
+                           true)?;
+
+    Ok(())
+}
+
+#[test]
+#[ntest::timeout(600000)]
+fn revoked_subkey() -> Result<()> {
+    let mut experiment = make_experiment!()?;
+    let cert = experiment.artifact(
+        "cert",
+        || -> Result<Cert> {
+            let cert = CertBuilder::new()
+                .set_creation_time(Experiment::now() - Duration::new(3600, 0))
+                .add_userid("Alice Lovelace <alice@lovelace.name>")
+                .add_signing_subkey()
+                .generate()
+                .map(|(cert, _rev)| cert)?;
+
+            let mut signer = cert.primary_key().key().clone()
+                .parts_into_secret()?.into_keypair()?;
+            let subkey = cert.keys().subkeys().nth(0).unwrap();
+            let sig = SubkeyRevocationBuilder::new()
+                .set_signature_creation_time(
+                    Experiment::now() - Duration::new(1800, 0))?
+                .set_reason_for_revocation(ReasonForRevocation::KeyRetired,
+                                           b"don't use it anymore")?
+                .build(&mut signer, &cert, subkey.key(), None)?;
+
+            Ok(cert.insert_packets2(sig)?.0)
+        },
+        |a, f| a.as_tsk().serialize(f),
+        |b| Cert::from_bytes(&b))?;
+
+    let sig = experiment.artifact(
+        "sig",
+        || -> Result<Packet> {
+            let mut signer = cert.keys().subkeys()
+                .nth(0).unwrap().key().clone()
+                .parts_into_secret()?.into_keypair()?;
+
+            SignatureBuilder::new(SignatureType::Binary)
+                .set_signature_creation_time(Experiment::now())?
+                .sign_message(&mut signer, MSG)
+                .map(Into::into)
+        },
+        |a, f| a.serialize(f),
+        |b| Packet::from_bytes(&b))?;
+
+    test_detached_sig_with(&mut experiment,
+                           true, &cert,
+                           sig, vec![],
+                           true,
+                           true)?;
+
+    Ok(())
+}
+
+#[test]
+#[ntest::timeout(600000)]
+fn revoked_cert_and_subkey() -> Result<()> {
+    let mut experiment = make_experiment!()?;
+    let cert = experiment.artifact(
+        "cert",
+        || -> Result<Cert> {
+            let (cert, rev) = CertBuilder::new()
+                .set_creation_time(Experiment::now() - Duration::new(3600, 0))
+                .add_userid("Alice Lovelace <alice@lovelace.name>")
+                .add_signing_subkey()
+                .generate()?;
+
+            let mut signer = cert.primary_key().key().clone()
+                .parts_into_secret()?.into_keypair()?;
+            let subkey = cert.keys().subkeys().nth(0).unwrap();
+            let sig = SubkeyRevocationBuilder::new()
+                .set_signature_creation_time(
+                    Experiment::now() - Duration::new(1800, 0))?
+                .set_reason_for_revocation(ReasonForRevocation::KeyRetired,
+                                           b"don't use it anymore")?
+                .build(&mut signer, &cert, subkey.key(), None)?;
+
+            Ok(cert.insert_packets2(vec![sig, rev])?.0)
+        },
+        |a, f| a.as_tsk().serialize(f),
+        |b| Cert::from_bytes(&b))?;
+
+    let sig = experiment.artifact(
+        "sig",
+        || -> Result<Packet> {
+            let mut signer = cert.keys().subkeys()
+                .nth(0).unwrap().key().clone()
+                .parts_into_secret()?.into_keypair()?;
+
+            SignatureBuilder::new(SignatureType::Binary)
+                .set_signature_creation_time(Experiment::now())?
+                .sign_message(&mut signer, MSG)
+                .map(Into::into)
+        },
+        |a, f| a.serialize(f),
+        |b| Packet::from_bytes(&b))?;
+
+    test_detached_sig_with(&mut experiment,
+                           true, &cert,
+                           sig, vec![],
+                           true,
+                           true)?;
+
+    Ok(())
+}
